// based on www.inb.uni-luebeck.de/~boehme/avcodec_sample.cpp
// This version was written by Michael Zeilfelder
#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>

#include <stdio.h>

#include "video.h"
#include "wx/wx.h"

VideoManager* VideoManager::mThis = NULL;

VideoManager* VideoManager::Create()
{
    if ( !mThis )
        mThis = new VideoManager();
    
    return mThis;
}

void VideoManager::Destroy()
{
    delete mThis;
    mThis = NULL;
}

VideoManager::VideoManager()
{
    // Register all formats and codecs
    av_register_all();
}

VideoManager::~VideoManager()
{
    av_free_static();
}

Video::Video()
    : mCodecCtx(NULL)
    , mCodec(NULL)
    , mVideoStream(-1)
    , mFrame(NULL)
    , mFrameRGB(NULL)
    , mFormatCtx(NULL)
    , mLastPacket(NULL)
    , mPosition(0)
    , mStartTime(0)
{
    mLastPacket = new AVPacket();
    memset(mLastPacket, 0, sizeof(AVPacket));
}

Video::~Video()
{
    CloseFile();
    delete mLastPacket;
}

void Video::CloseFile()
{
    if ( mFrameRGB )
        av_free(mFrameRGB);
    mFrameRGB = NULL;

    if ( mFrame )
        av_free(mFrame);
    mFrame = NULL;
    
    if ( mCodecCtx )
        avcodec_close(mCodecCtx);
    mCodecCtx = NULL;
    
    if ( mFormatCtx )
        av_close_input_file(mFormatCtx);    
    mFormatCtx = NULL;
    
    mVideoStream = -1;
}

bool Video::LoadFile(const char* filename_)
{
    CloseFile();
    
    // Open video file
    if ( av_open_input_file(&mFormatCtx, filename_, NULL, 0, NULL) != 0 )
        return false; // Couldn't open file
    
    // Retrieve stream information
    if ( av_find_stream_info(mFormatCtx) < 0 )
    {
        fprintf(stderr, "Couldn't find stream information\n");
        CloseFile();
        return false;
    }
    
    // Find the first video stream
    mVideoStream=-1;
    for ( int i=0; i<mFormatCtx->nb_streams; i++ )
    {
        if ( mFormatCtx->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO )
        {
            mVideoStream = i;
            break;
        }
    }
    
    if ( mVideoStream==-1 )
    {
        fprintf(stderr, "Didn't find a video stream\n");
        CloseFile();
        return false;
    }
    
    // Get a pointer to the codec context for the video stream
    mCodecCtx = mFormatCtx->streams[mVideoStream]->codec;
    
    // Find the decoder for the video stream
    mCodec = avcodec_find_decoder(mCodecCtx->codec_id);
    if ( mCodec == NULL )
    {
        fprintf(stderr, "Codec not found\n");
        mCodecCtx = NULL;
        CloseFile();
        return false;
    }
    
    mCodecCtx->workaround_bugs = 1;    // not sure what this does, i just found it in the ffplaysample and it sounded useful
    if ( mFormatCtx->start_time != AV_NOPTS_VALUE )
        mStartTime = mFormatCtx->start_time;
    
    // Open codec
    if ( avcodec_open(mCodecCtx, mCodec) < 0 )
    {
        fprintf(stderr, "Could not open codec\n");
        mCodecCtx = NULL;
        CloseFile();
        return false;
    }
    
    // Hack to correct wrong frame rates that seem to be generated by some 
    // codecs
//    if (    mCodecCtx->frame_rate > 1000 
  //       && mCodecCtx->frame_rate_base == 1 )
    //    mCodecCtx->frame_rate_base = 1000;
    
    // Allocate video frame
    mFrame = avcodec_alloc_frame();
    
    // Allocate an AVFrame structure
    mFrameRGB = avcodec_alloc_frame();
    
    if(!mFrame || !mFrameRGB )
    {
        fprintf(stderr, "not enough memory\n");
        CloseFile();
        return false;
    }
    
    return true;
}

void Video::GetSize(int & width_, int &height_)
{
    if ( mCodecCtx )
    {
        width_ = mCodecCtx->width;
        height_ = mCodecCtx->height;
    }
}

bool Video::GetNextFrame( wxImage * targetImage_ )
{
    if ( mVideoStream < 0 )
    {
        fprintf(stderr, "no video stream opened\n");
        return false;
    }
    
    if ( !targetImage_ )
        return false;
    
    if (    targetImage_->GetWidth() != mCodecCtx->width 
         || targetImage_->GetHeight() != mCodecCtx->height )
    {
        fprintf(stderr, "target image has wrong size\n");
        return false;
    }
    
    int frameFinished = 0;
    
    unsigned char * data = targetImage_->GetData();
    if ( !data )
        return false;
    
    // Assign appropriate parts of buffer to image planes in pFrameRGB
    if ( avpicture_fill((AVPicture *)mFrameRGB, data, PIX_FMT_RGB24,
         mCodecCtx->width, mCodecCtx->height) < 0 )
    {
        fprintf(stderr, "avpicture_fill failed\n");
        return false;
    }
    
    
    while ( av_read_frame(mFormatCtx, mLastPacket) >= 0 ) 
    {   
        // Is this a packet from the video stream?
        if ( mLastPacket->stream_index == mVideoStream )
        {
            // Decode video frame
            avcodec_decode_video( mCodecCtx, mFrame, &frameFinished, 
                                  mLastPacket->data, mLastPacket->size);
         }

        // Free the packet that was allocated by av_read_frame
        av_free_packet(mLastPacket);

        if ( frameFinished )
            break;
    }
    
    if ( frameFinished )
    {
        ++ mPosition;
        // fprintf(stderr, "mPosition:%ld frame_number:%d ", mPosition, mCodecCtx->frame_number );
        
        // Convert the image from its native format to RGB
        if ( img_convert( (AVPicture *)mFrameRGB, 
                     PIX_FMT_RGB24,
                     (AVPicture*)mFrame, 
                     mCodecCtx->pix_fmt, 
                     mCodecCtx->width, 
                     mCodecCtx->height ) < 0 )
        {
            fprintf(stderr, "img_convert failed\n");
            return false;
        }
        
        return true;
    }
    else
    {
        //fprintf(stderr, "av_read_frame failed\n");
    }
    
    // fprintf(stderr, "GetNextFrame failed\n");
    return false;
}

void Video::DebugStreamInfo( AVStream * stream_)
{
    if ( !stream_ )
        return;
    
    fprintf(stderr, " *** Debuginfo AVStream ***\n");
   // fprintf(stderr, "index %d\n", stream_->index);
//    fprintf(stderr, "id %d\n", stream_->id);
    fprintf(stderr, "r_frame_rate %d\n", stream_->r_frame_rate);
//    fprintf(stderr, "r_frame_rate_base %d\n", stream_->r_frame_rate_base);
//    fprintf(stderr, "codec_info_duration %ld\n", stream_->codec_info_duration);
//    fprintf(stderr, "codec_info_nb_frames %d\n", stream_->codec_info_nb_frames);
//    fprintf(stderr, "pts_wrap_bits %d\n", stream_->pts_wrap_bits);
//    fprintf(stderr, "quality %f\n", stream_->quality);
    fprintf(stderr, "start_time %ld\n", stream_->start_time);
    fprintf(stderr, "duration %ld\n", stream_->duration);
//    fprintf(stderr, "need_parsing %d\n", stream_->need_parsing);
    fprintf(stderr, "cur_dts %ld\n", stream_->cur_dts);
//    fprintf(stderr, "last_IP_duration %d\n", stream_->last_IP_duration);
//    fprintf(stderr, "last_IP_pts %ld\n", stream_->last_IP_pts);
//    fprintf(stderr, "nb_index_entries %d\n", stream_->nb_index_entries);
//    fprintf(stderr, "index_entries_allocated_size %d\n", stream_->index_entries_allocated_size);
}

bool Video::Seek(long frame_)
{
//    fprintf(stderr, "seek frame %ld \n", frame_);
    if ( mVideoStream < 0 )
    {
        fprintf(stderr, "no video stream opened\n");
        return false;
    }

    int64_t pos = FrameToTimeUnits(frame_);
    if ( av_seek_frame(mFormatCtx, -1, mStartTime + pos, 0 ) < 0 )
    {
        fprintf(stderr, "av_seek_frame failed \n");
        return false;
    }
    mPosition = frame_;
    avcodec_flush_buffers(mCodecCtx);    
    
    return true;
}

long Video::GetDuration()
{
    if ( !mFormatCtx )
        return -1;
    
    return (long)TimeUnitsToFrame(mFormatCtx->duration);
}

long Video::GetPosition()
{
    return mPosition;
}

double Video::GetFrameRate()
{
    if ( !mCodecCtx )
        return 25.f;


    // Determine the fps
    double framerate = ( double )mCodecCtx->time_base.den / ( mCodecCtx->time_base.num == 0 ? 1 : mCodecCtx->time_base.num );

    // We'll use fps if it's available
    if ( framerate > 0 && framerate < 30 )
        return framerate;

    return 25.f;
}

int64_t Video::TimeUnitsToFrame(int64_t units_)
{
    return (int64_t)((units_*GetFrameRate()) / AV_TIME_BASE);
}

int64_t Video::FrameToTimeUnits(int64_t frame_)
{
    return (int64_t)((frame_*AV_TIME_BASE)/GetFrameRate());
}
